from pymongo import MongoClient, ASCENDING
import time


class MongoCollection(object):
    """This class represents a collection in MongoDB. It contains a number
    of useful utilities, like index management and workload execution
    on that collection.
    """

    def __init__(self, uri, db, coll):
        self.collection_name = coll
        self.namespace = f"{db}.{coll}"
        self.uri = uri
        self.client = MongoClient(uri)
        self.db = self.client[db]
        self.collection = self.client[db][coll]
        self.count = self.collection.count_documents({})
        self.last_index = None

    def _get_query_options(self, query):
        """returns sort, limit and projection of a query in dictionary form."""
        if isinstance(query, dict):
            return {"sort": None, "limit": 0, "projection": None}

        sort = [(field, 1) for field in query.sort] if query.sort else None
        limit = query.limit or 0
        if query.projection:
            projection = {field: 1 for field in query.projection}
            # explicitly exclude _id
            projection["_id"] = 0
        else:
            projection = None

        return {"sort": sort, "limit": limit, "projection": projection}

    def _parse_winning_plan(self, explain):
        stages = []
        doc = explain["executionStages"]
        while "stage" in doc:
            stages.append(doc["stage"])
            doc = doc.get("inputStage", {})
        stages.reverse()
        return " -> ".join(stages)

    def execute_query(self, query):
        """Executes the query with find syntax and exhausts the cursor.
        query can be a Query instance or a MQL query as a dictionary. Note: this does
        not return the results, use the collection instance member instead if you need
        the results of the query.
        """
        options = self._get_query_options(query)
        options["allow_disk_use"] = True
        cursor = self.collection.find(query.to_mql(), **options)
        for _ in cursor:
            pass

    def explain_query(self, query):
        """Explains a single query in executionStats mode for workload measurements.
        query can be a Query instance or a MQL query as a dictionary."""
        options = self._get_query_options(query)

        command = {
            "explain": {
                "find": self.collection_name,
                "filter": query.to_mql(),
                "limit": options["limit"],
                "projection": options["projection"] or {},
                "sort": dict(options["sort"]) if options["sort"] else {},
                "allowDiskUse": True,
            },
            "verbosity": "executionStats",
        }
        result = self.db.command(command)
        return result["executionStats"]

    def execute_workload(self, workload, explain=True):
        """executes a query workload and returns the time in milliseconds it took."""
        exec_time = 0
        t = time.time()
        for i, query in enumerate(workload):
            if explain:
                result = self.explain_query(query)
                print(
                    f"{i:2}  {query}\n"
                    f"     executionTimeMillis {result['executionTimeMillis']}"
                    f"     totalKeysExamined {result['totalKeysExamined']}"
                    f"     totalDocsExamined {result['totalDocsExamined']}"
                    f"     nReturned {result['nReturned']}"
                    f"     plan {self._parse_winning_plan(result)}"
                )

                exec_time += result["executionTimeMillis"]
            else:
                self.execute_query(query)
        if explain:
            return exec_time
        return (time.time() - t) * 1000

    def list_indexes(self):
        """Lists all available indexes on the collection by name."""
        return [idx.get("name") for idx in self.collection.list_indexes()]

    def create_index(self, index):
        """create an index on the collection. index is a field tuple. e.g. ("foo", "bar")
        which would create the index {"foo": 1, "bar": 1}.
        """
        asc = [ASCENDING] * len(index)
        index_name = "_".join(index).replace(" ", "_")
        index = list(zip(index, asc))

        self.last_index = index_name
        self.collection.create_index(index, name=index_name)

    def drop_last_index(self):
        """drops the last index created with create_index(). Only one index can be
        removed this way (not a stack).
        """
        if self.last_index:
            self.collection.drop_index(self.last_index)
            self.last_index = None

    def drop_indexes(self):
        """drops all indexes of the given collection."""
        self.last_index = None
        self.collection.drop_indexes()



from .common import map_bson, QueryRegionEmptyException
import math

import pandas as pd

SAMPLE_DB_NAME = "samples"

# Get a cardinality estimation via sampling
class SampleEstimator:
    def __init__(
        self,
        mongo_collection,
        numrows=None,
        sample_ratio=None,
        sample_size=None,
        persist=False,
        sample_db_name=SAMPLE_DB_NAME,
        **kwargs,
    ):
        # mongoCollection is a MongoCollection class object used to access collection
        self.mongo = mongo_collection
        self.columns = None
        self.sample_size = sample_size
        self.sample_db_name = sample_db_name
        self.kwargs = kwargs

        self.limit = None
        if numrows:
            if numrows < self.mongo.count:
                self.limit = numrows

        db_size = self.limit or self.mongo.count

        self.persist = persist

        if sample_ratio is not None:
            assert 0 < sample_ratio <= 1, "sample_ratio must be between 0 and 1"

            assert (
                sample_size == None
            ), "please provide only one: sample_size or sample_ratio"

            self.sample_size = round(db_size * sample_ratio)

        if self.sample_size is not None:
            assert (
                0 < self.sample_size <= db_size
            ), "sample size must be between 0 and the size of the collection"

        if self.sample_size == db_size:
            self.sample_size = None  # Remove the sample stage if it's equal to the size of the collection and do not persist
            self.persist = False

        if persist:
            # create and save a sample collection in the samples database with sample_size = self.sample_size
            pipeline = [
                {"$sample": {"size": self.sample_size}},
                {
                    "$out": {
                        "db": self.sample_db_name,
                        "coll": self.mongo.collection_name,
                    }
                },
            ]
            if self.limit:
                pipeline.insert(0, {"$limit": self.limit})

            self.mongo.collection.aggregate(pipeline, allowDiskUse=True)

    def get_cardinality(self):
        return self.limit or self.mongo.count

    def make_pipeline(self, query):
        pipeline = [
            {"$match": query.to_mql()},
            {"$count": "total"},
        ]

        # take a sample from the actual database if a persistant sample was not created
        if self.sample_size and not self.persist:
            pipeline.insert(0, {"$sample": {"size": self.sample_size}})

        if self.limit and not self.persist:
            pipeline.insert(0, {"$limit": self.limit})

        return pipeline

    def estimate(self, query):
        """Returns the cardinality of a query by executing the query on the
        MongoDB collection and returning the result.
        """
        pipeline = self.make_pipeline(query)

        try:
            # if persistant sample has been created, query on that collection in db samples instead
            if self.persist:
                coll = self.mongo.collection_name
                actual = (
                    self.mongo.client[self.sample_db_name][coll]
                    .aggregate(pipeline, allowDiskUse=True)
                    .next()["total"]
                )
            else:
                # otherwise query on the full collection
                actual = self.mongo.collection.aggregate(
                    pipeline, allowDiskUse=True
                ).next()["total"]

            if self.sample_size:
                # if a sample size was provided, multiply the result to estimate the cardinality of the entire db
                total_count = self.limit or self.mongo.count
                actual = round(actual * (total_count / self.sample_size))

            return actual
        except:
            return 0

    def drop_sample(self):
        # drops any persisted samples in the samples database
        coll_name = self.mongo.collection_name
        return self.mongo.client[self.sample_db_name][coll_name].drop()

    def sample(self, query, n):
        pipeline = [
            {"$match": query.to_mql()},
            {"$sample": {"size": n}},
        ]

        try:
            if self.persist:
                coll = self.mongo.collection_name
                cur = self.mongo.client[self.sample_db_name][coll].aggregate(
                    pipeline, allowDiskUse=True
                )
            else:
                cur = self.mongo.collection.aggregate(pipeline, allowDiskUse=True)
        except StopIteration:
            raise Exception("Query region is empty")

        df = pd.json_normalize(cur)
        for col in df.columns:
            df[col] = df[col].map(map_bson)

        # if the query matches no documents in the database
        if len(df) == 0:
            raise QueryRegionEmptyException(
                "Error in SampleEstimator: Query region is empty"
            )

        # if not enough results are returned
        if len(df) < n:
            duplicate = math.ceil(n / len(df))
            # duplicate the existing data until we have n results
            df = df.append([df] * duplicate, ignore_index=True)
            df = df.head(n)

        # shuffle the result returned
        df = df.sample(frac=1).reset_index(drop=True)

        return df



from typing import Tuple
from .common import OPERATORS

from collections import namedtuple
import numpy as np
import pandas as pd

Predicate = namedtuple("Predicate", ["column", "op", "values"])


class Query(object):
    """The Query class represents a query object, consisting of
    a number of Predicates. Predicates are named tuples of
    the form (column, op, values).
    This class also contains some utility methods, for example
    exporting a query to MQL syntax, intersecting a query
    with an index (for Index Selection), etc.
    """

    def __init__(self):
        self.predicates = []

        self._limit = None
        self._projection = None
        self._sort = None

    @property
    def limit(self):
        """returns the limit for this query."""
        return self._limit

    @limit.setter
    def limit(self, n: int):
        """set a limit for this query. None means no limit."""
        assert type(n) == int, "limit must be an integer"
        self._limit = n

    @property
    def projection(self):
        """returns the projection tuple of this query."""
        return self._projection

    @projection.setter
    def projection(self, p: Tuple[str, ...]):
        """set the projection tuple for this query."""
        self._projection = p

    @property
    def sort(self):
        """returns the sort tuple of this query."""
        return self._sort

    @sort.setter
    def sort(self, s: Tuple[str, ...]):
        """set the sort tuple for this query."""
        self._sort = s

    @property
    def fields(self):
        """return all fields of the query, whether they are part of the predicates,
        projection or sort.
        """
        fields = [pred.column for pred in self.predicates]
        if self.sort:
            fields += list(self.sort)
        if self.projection:
            fields += list(self.projection)
        return sorted(fields)

    def add_predicate(self, predicate):
        """adds a predicate to the query."""

        assert isinstance(predicate.values, list)
        assert (
            predicate.op in OPERATORS.keys()
        ), f"unsupported operator '${predicate.op}'"

        # simplify $in
        if predicate.op in ["in", "nin"]:
            assert len(predicate.values) >= 1, "in operator requires at least 1 value"
            if len(predicate.values) == 1:
                predicate.op == "eq" if predicate.op == "in" else "neq"
        else:
            assert len(predicate.values) == 1, "operator takes exactly 1 value"

        self.predicates.append(predicate)

    def add_predicates(self, preds):
        """add multiple predicates to a query at once."""
        for pred in preds:
            self.add_predicate(pred)

    def index_intersect(self, index):
        """returns a copy of this query that only contains predicates on the
        provided index fields, left to right, up until an index field was
        not included in the query.

        Example: query is {a: true, b: {$lt: 20}, c: 5}
                 query.index_intersect(("b", "d", "c")) return the query
                 equivalent to {b: {$lt: 20}}, because d is not present
                 in the query and aborts the algorithm.

        """
        query = Query()

        for field in index:
            # reset field indicator
            query_has_field = False
            for pred in self.predicates:
                if pred.column == field:
                    query_has_field = True
                    query.add_predicate(pred)
            # if no predicates found for this field, stop.
            if not query_has_field:
                break
        return query

    def is_subset(self, index):
        """returns true if all predicate fields are included in the index.
        This is necessary, but not sufficient to be a covered by the index.
        It is used to determine if a limit caps the cost of the query or not.
        """
        predicate_fields = set(p.column for p in self.predicates)
        return predicate_fields.issubset(set(index))

    def is_covered(self, index):
        """returns true if this query is covered by the index, false otherwise.
        A query without a projection is never covered. A query with projection
        is covered if the union of all predicate fields and projected fields
        appear in the index.
        """
        if self.projection is None:
            return False

        predicate_fields = tuple(p.column for p in self.predicates)
        fields_to_cover = set(predicate_fields + self.projection)

        return fields_to_cover.issubset(set(index))

    def can_use_sort(self, index):
        """returns true if this query has a sort and the index can be used to sort,
        false otherwise.
        A query can be sorted by an index if any of the following is true:
            a) the sort fields are the same as the index fields (incl. order)
            b) the sort fields are a prefix sequence of the index fields
            c) the sort fields are a sub-sequence of the index fields and
               the query only uses equality predicates on all fields preceeding
               the sort fields.

               Example: index on ('a', 'b', 'c', 'd')
                        query {a: 5, b: {$gt: 6}} with sort ('b', 'c')
                        The predicate preceeding the sort fields is 'a' and is
                        an equality predicate. The index can be used to sort.
        """
        if self.sort is None:
            # TODO check this: if the query doesn't need a sort, the sort fields sequence
            # is and empty list, and theoretically a prefix of any index.
            return True

        # cover case a) and b)
        if self.sort == index[: len(self.sort)]:
            return True

        # cover case c)
        sub_idx = -1
        for i in range(len(index) - len(self.sort)):
            if self.sort == index[i : i + len(self.sort)]:
                # sort fields are a sub-sequence of index fields
                sub_idx = i

        if sub_idx != -1:
            # check for preceeding equality predicates
            if all(pred.op == "eq" for pred in self.predicates[:sub_idx]):
                return True

        return False

    def __repr__(self):
        s = []
        s.append(f"filter={self.to_mql()}")
        if self.sort:
            s.append(f"sort={self.sort}")
        if self.limit:
            s.append(f"limit={self.limit}")
        if self.projection:
            s.append(f"projection={self.projection}")
        return f"Query({', '.join(s)})"

    def __len__(self):
        """The length of a query, using 'len(query)', is the number of
        predicates it contains."""
        return len(self.predicates)

    def to_df_query(self, df):
        """Converts the query to be used on a Pandas DataFrame."""
        FUNC_MAP = {"lt": "lt", "lte": "le", "gt": "gt", "gte": "ge", "eq": "eq"}
        bools = pd.Series([True] * df.shape[0])
        for pred in self.predicates:
            fn = FUNC_MAP[pred.op]
            colname = pred.column.replace("_", " ")
            newbools = getattr(df[colname], fn)(pred.values[0])
            bools &= newbools
        return bools.sum()

    def to_mql(self):
        """converts the query to MQL syntax."""
        query = {}

        def clean_value(v):
            if type(v) == str:
                return v.strip()
            elif isinstance(v, np.int64):
                return v.item()
            else:
                return v

        for predicate in self.predicates:
            name = predicate.column
            # remove leading/trailing whitespace
            values = [clean_value(v) for v in predicate.values]
            if predicate.op == "eq":
                if pd.isna(values[0]):
                    # special handling for missing values
                    query[name] = {"$exists": False}
                else:
                    query[name] = values[0]
            elif predicate.op in ["in", "nin"]:
                query[name] = {f"${predicate.op}": values}
            else:
                if name not in query:
                    query[name] = {f"${predicate.op}": values[0]}
                else:
                    if isinstance(query[name], dict):
                        query[name][f"${predicate.op}"] = values[0]
                    else:
                        query[name] = {
                            "$eq": query[name],
                            f"${predicate.op}": values[0],
                        }
        return query

    @classmethod
    def from_mql(cls, query):
        """converts MQL syntax to a Query object."""
        q = cls()

        if not query:
            return q

        for col, value in query.items():
            if col.startswith("$"):
                raise AssertionError(f"unsupported operator '{col}'")
            if isinstance(value, (str, int)):
                q.add_predicate(Predicate(column=col, op="eq", values=[value]))
            elif isinstance(value, dict):
                ops = value.keys()
                for item in ops:
                    if item in ["$in", "$nin"]:
                        q.add_predicate(
                            Predicate(
                                column=col, op=item.lstrip("$"), values=value[item]
                            )
                        )
                    elif item == "$exists" and not value[item]:
                        # special handling for missing values
                        q.add_predicate(Predicate(column=col, op="eq", values=[None]))
                    else:
                        q.add_predicate(
                            Predicate(
                                column=col, op=item.lstrip("$"), values=[value[item]]
                            )
                        )

        return q


if __name__ == "__main__":

    # Example Usage of a Query objet
    query = Query()

    query.add_predicates(
        [
            Predicate("Suspension_Indicator", "eq", ["Y"]),
            Predicate("Make", "in", ["INFIN", "HYUND"]),
        ]
    )



import numpy as np
import pandas as pd
import operator

from bson.timestamp import Timestamp
from bson.int64 import Int64
from bson.decimal128 import Decimal128

from datetime import datetime

LOG2PI = np.log(2 * np.pi)

MISSING_TYPE = "_missing_"
TYPE_COL_NAME = "_types_"


def try_compare(x, op, y, default=False):
    try:
        return op(x, y)
    except TypeError as t:
        # if the types are not comparable, it's not a match
        # MongoDB calls this type bracketing
        return default


def eq_with_nan(a, b):
    if pd.isna(a) and pd.isna(b):
        return True
    return operator.eq(a, b)


OPERATORS = {
    "gt": lambda a, val: try_compare(a, operator.gt, val[0]),
    "lt": lambda a, val: try_compare(a, operator.lt, val[0]),
    "gte": lambda a, val: try_compare(a, operator.ge, val[0]),
    "lte": lambda a, val: try_compare(a, operator.le, val[0]),
    "eq": lambda a, val: try_compare(a, eq_with_nan, val[0]),
    "ne": lambda a, val: try_compare(a, operator.ne, val[0], True),
    "in": lambda a, val: try_compare(val, operator.contains, a),
    "nin": lambda a, val: try_compare(
        val, lambda x, y: operator.not_(operator.contains(x, y)), a, True
    ),
}


def get_type_or_missing(x):
    if pd.isna(x) or x == None:
        return MISSING_TYPE
    return type(x).__name__


def map_bson(b):
    if isinstance(b, (bool, int, float, str, datetime)):
        return b
    elif isinstance(b, Timestamp):
        return b.as_datetime()
    elif isinstance(b, Int64):
        return int(str(b))
    elif isinstance(b, Decimal128):
        return b.to_decimal()
    elif isinstance(b, list):
        raise TypeError("We don't know how to map arrays, yet")
    return str(b)


class QueryRegionEmptyException(Exception):
    pass



import pickle


class Workload(object):
    """Represents a workload consisting of a number of Query objects.
    In its current form, a workload is defined by a number of
    query patterns (= shapes), and a number of queries of the
    given shape. The workload is randomly generated based on these
    parameters.
    """

    def __init__(self):
        self.queries = []

    def __repr__(self):
        return f"<Workload: {len(self.queries)} queries>"

    def __iter__(self):
        return iter(self.queries)

    def __len__(self):
        return len(self.queries)

    def print(self):
        """Prints each query of the workload."""
        for i, query in enumerate(self.queries):
            print(f"{i:>4} {query}")

    def save(self, filename):
        """Save a workload to a file (without the model)"""
        # save workload
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        """Class method to load a workload from file."""
        # read workload from file
        with open(filename, "rb") as f:
            workload = pickle.load(f)

        return workload
